import os
import random
import string

from future.moves.urllib.parse import urlsplit

from mendeley.auth import MendeleyClientCredentialsAuthenticator, \
    MendeleyAuthorizationCodeAuthenticator, \
    MendeleyImplicitGrantAuthenticator


class Mendeley(object):
    def __init__(self,
                 client_id,
                 client_secret=None,
                 redirect_uri=None,
                 host='https://api.mendeley.com',
                 state_generator=None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.host = host
        self.state_generator = state_generator or DefaultStateGenerator()

        if is_localhost(redirect_uri):
            os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1'

    def start_client_credentials_flow(self):
        return MendeleyClientCredentialsAuthenticator(self)

    def start_authorization_code_flow(self, state=None):
        state = state or self.state_generator.generate_state()
        return MendeleyAuthorizationCodeAuthenticator(self, state)

    def start_implicit_grant_flow(self, state=None):
        state = state or self.state_generator.generate_state()
        return MendeleyImplicitGrantAuthenticator(self, state)


class DefaultStateGenerator(object):
    ASCII_CHARACTER_SET = string.ascii_uppercase + string.digits

    @staticmethod
    def generate_state(length=30, chars=ASCII_CHARACTER_SET):
        rand = random.SystemRandom()
        return ''.join(rand.choice(chars) for _ in range(length))


def is_localhost(url):
    if not url:
        return False

    split_url = urlsplit(url)
    return split_url.scheme == 'http' and split_url.hostname in ['127.0.0.1', '0.0.0.0', 'localhost']



from memoized_property import memoized_property


class Page(object):
    """
    A page of a collection of objects.
    """
    def __init__(self, session, rsp, obj_type, count=None):
        self.session = session
        self.rsp = rsp
        self.obj_type = obj_type

        if count:
            self.count = count
        elif 'mendeley-count' in self.rsp.headers:
            self.count = int(self.rsp.headers.get('mendeley-count'))
        else:
            self.count = len(self.items)

    @memoized_property
    def items(self):
        """
        a list of items on this page.
        """
        return [self.obj_type(self.session, i) for i in self.rsp.json()]

    def _navigate(self, rel):
        if rel not in self.rsp.links:
            return None
        rsp = self.session.get(self.rsp.links[rel]['url'])
        return Page(self.session, rsp, self.obj_type, self.count)

    @property
    def first_page(self):
        """
        the first :class:`Page <mendeley.pagination.Page>` of the collection.
        """
        return self._navigate('first')

    @property
    def last_page(self):
        """
        the first :class:`Page <mendeley.pagination.Page>` of the collection.
        """
        return self._navigate('next')

    @property
    def next_page(self):
        """
        the next :class:`Page <mendeley.pagination.Page>` of the collection, or `None` if this is the last page.
        """
        return self._navigate('next')

    @property
    def previous_page(self):
        """
        the previous :class:`Page <mendeley.pagination.Page>` of the collection, or `None` if this is the first page.
        """
        return self._navigate('prev')



from mendeley.models.files import File
from mendeley.resources.base import ListResource


class Files(ListResource):
    """
    Top-level resource for accessing files.  These can be:

    - files for the logged-in user, if retrieved from a
      :func:`MendeleySession <mendeley.session.MendeleySession.files>`.
    - files attached to a :func:`CatalogDocument <mendeley.models.catalog.CatalogDocument.files>`.
    - files attached to a :func:`UserDocument <mendeley.models.documents.UserDocument.files>`.
    - files in a :func:`Group <mendeley.models.groups.Group.files>`.
    """
    _url = '/files'

    def __init__(self, session, catalog_id=None, document_id=None, group_id=None):
        self.session = session
        self.catalog_id = catalog_id
        self.document_id = document_id
        self.group_id = group_id

    def get(self, id):
        """
        Retrieves a file by ID.

        :param id: the ID of the file to get.
        :return: a :class:`File <mendeley.models.files.File>`.
        """
        return super(Files, self).get(id)

    def list(self, page_size=None, added_since=None, deleted_since=None):
        """
        Retrieves files, as a paginated collection.

        :param page_size: the number of files to return on each page.  Defaults to 20.
        :param added_since: if specified, only returns files added after this timestamp.
        :param deleted_since: if specified, only returns the IDs of files deleted after this timestamp.
        :return: a :class:`Page <mendeley.pagination.Page>` of :class:`Files <mendeley.models.files.File>`.
        """
        return super(Files, self).list(page_size,
                                       added_since=added_since,
                                       deleted_since=deleted_since,
                                       catalog_id=self.catalog_id,
                                       document_id=self.document_id,
                                       group_id=self.group_id)

    def iter(self, page_size=None, added_since=None, deleted_since=None):
        """
        Retrieves files, as an iterator.

        :param page_size: the number of files to retrieve at a time.  Defaults to 20.
        :param added_since: if specified, only returns files added after this timestamp.
        :param deleted_since: if specified, only returns the IDs of files deleted after this timestamp.
        :return: an iterator of :class:`Files <mendeley.models.files.File>`.
        """
        return super(Files, self).iter(page_size,
                                       added_since=added_since,
                                       deleted_since=deleted_since,
                                       catalog_id=self.catalog_id,
                                       document_id=self.document_id,
                                       group_id=self.group_id)

    @property
    def _session(self):
        return self.session

    def _obj_type(self, **kwargs):
        return File


import arrow

from mendeley.models.common import Discipline, Photo, Location, Education, Employment
from mendeley.response import SessionResponseObject


class Profile(SessionResponseObject):
    """
    A Mendeley profile.

    .. attribute:: id
    .. attribute:: first_name
    .. attribute:: last_name
    .. attribute:: display_name
    .. attribute:: email
    .. attribute:: link
    .. attribute:: research_interests
    .. attribute:: academic_status
    .. attribute:: verified
    .. attribute:: user_type
    """
    content_type = 'application/vnd.mendeley-profiles.1+json'

    @property
    def created(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'created' in self.json:
            return arrow.get(self.json['created'])
        else:
            return None

    @property
    def discipline(self):
        """
        a :class:`Discipline <mendeley.models.common.Discipline>`.
        """
        if 'discipline' in self.json:
            return Discipline(self.json['discipline'])
        else:
            return None

    @property
    def photo(self):
        """
        a :class:`Photo <mendeley.models.common.Photo>`.
        """
        if 'photo' in self.json:
            return Photo(self.json['photo'])
        else:
            return None

    @property
    def location(self):
        """
        a :class:`Location <mendeley.models.common.Location>`.
        """
        if 'location' in self.json:
            return Location(self.json['location'])
        else:
            return None

    @property
    def education(self):
        """
        a list of :class:`Education <mendeley.models.common.Education>` objects.
        """
        if 'education' in self.json:
            return [Education(e) for e in self.json['education']]
        else:
            return None

    @property
    def employment(self):
        """
        a list of :class:`Employment <mendeley.models.common.Employment>` objects.
        """
        if 'employment' in self.json:
            return [Employment(e) for e in self.json['employment']]
        else:
            return None

    @classmethod
    def fields(cls):
        return ['id', 'first_name', 'last_name', 'display_name', 'email', 'link', 'research_interests',
                'academic_status', 'verified', 'user_type']



class ResponseObject(object):
    def __init__(self, json):
        self.json = json

    def __getattr__(self, name):
        if name in self.fields():
            return self.json.get(name)
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

    @classmethod
    def __dir__(cls):
        d = set(dir(cls) + cls.fields())
        d.remove('fields')

        return sorted(d)


class SessionResponseObject(ResponseObject):
    def __init__(self, session, json):
        super(SessionResponseObject, self).__init__(json)

        self.session = session


class LazyResponseObject(object):
    def __init__(self, session, id, obj_type, loader):
        self.session = session
        self.id = id

        self._obj_type = obj_type
        self._loader = loader

        self._value = None

    def __getattr__(self, name):
        return getattr(self._delegate, name)

    @property
    def _delegate(self):
        if not self._value:
            self._value = self._loader()

        return self._value

    def __dir__(self):
        d = set(dir(self.__class__) + self._obj_type.__dir__())

        return sorted(d)



import arrow

from mendeley.models.common import Photo
from mendeley.models.profiles import Profile
from mendeley.response import SessionResponseObject, LazyResponseObject


class Group(SessionResponseObject):
    """
    A Mendeley group.

    .. attribute:: id
    .. attribute:: name
    .. attribute:: description
    .. attribute:: disciplines
    .. attribute:: tags
    .. attribute:: webpage
    .. attribute:: link
    .. attribute:: access_level
    .. attribute:: role
    """
    content_type = 'application/vnd.mendeley-group.1+json'

    @property
    def created(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'created' in self.json:
            return arrow.get(self.json['created'])
        else:
            return None

    @property
    def photo(self):
        """
        a :class:`Photo <mendeley.models.common.Photo>`.
        """
        if 'photo' in self.json:
            return Photo(self.json['photo'])
        else:
            return None

    @property
    def owner(self):
        """
        a :class:`Profile <mendeley.models.profiles.Profile>`.
        """
        if 'owning_profile_id' in self.json:
            return self.session.profiles.get_lazy(self.json['owning_profile_id'])
        else:
            return None

    @property
    def members(self):
        return self.session.group_members(self.id)

    @property
    def documents(self):
        """
        a :class:`Documents <mendeley.resources.documents.Documents>` resource, from which
        :class:`UserDocuments <mendeley.models.documents.UserDocument>` can be retrieved.
        """
        return self.session.group_documents(self.id)

    @property
    def trash(self):
        """
        a :class:`Trash <mendeley.resources.trash.Trash>` resource, from which
        :class:`TrashDocuments <mendeley.models.documents.TrashDocument>` can be retrieved.
        """
        return self.session.group_trash(self.id)

    @property
    def files(self):
        """
        a :class:`Files <mendeley.resources.files.Files>` resource, from which
        :class:`Files <mendeley.models.files.File>` can be retrieved.
        """
        return self.session.group_files(self.id)

    @classmethod
    def fields(cls):
        return ['id', 'name', 'description', 'disciplines', 'tags', 'webpage', 'link', 'access_level',
                'role']


class GroupMember(LazyResponseObject, Profile):
    """
    A member of a Mendeley group.

    .. attribute:: id
    .. attribute:: role
    .. attribute:: first_name
    .. attribute:: last_name
    .. attribute:: display_name
    .. attribute:: email
    .. attribute:: link
    .. attribute:: research_interests
    .. attribute:: academic_status
    .. attribute:: verified
    .. attribute:: user_type
    """
    content_type = 'application/vnd.mendeley-membership.1+json'

    def __init__(self, session, member_json):
        super(GroupMember, self).__init__(session, member_json.get('profile_id'), Profile, lambda: self._load())

        self.member_json = member_json

    @property
    def joined(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'joined' in self.member_json:
            return arrow.get(self.member_json['joined'])
        else:
            return None

    @property
    def role(self):
        return self.member_json.get('role')

    def _load(self):
        return self.session.profiles.get(self.id)



import platform

from future.moves.urllib.parse import urljoin
from requests_oauthlib import OAuth2Session

from mendeley.exception import MendeleyApiException
from mendeley.resources import *
from mendeley.version import __version__


class MendeleySession(OAuth2Session):
    """
    Entry point for accessing Mendeley resources.

    .. attribute:: catalog

       A :class:`Catalog <mendeley.resources.catalog.Catalog>` resource for accessing the Mendeley catalog.

    .. attribute:: documents

       A :class:`Documents <mendeley.resources.documents.Documents>` resource for accessing documents in the logged-in
       user's library.

    .. attribute:: files

       A :class:`Files <mendeley.resources.files.Files>` resource for accessing files in the logged-in user's library.

    .. attribute:: groups

       A :class:`Groups <mendeley.resources.groups.Groups>` resource for accessing groups that the user is a member of.

    .. attribute:: profiles

       A :class:`Profiles <mendeley.resources.profiles.Profiles>` resource for accessing profiles of Mendeley users.

    .. attribute:: trash

       A :class:`Trash <mendeley.resources.trash.Trash>` resource for accessing trashed documents in the logged-in
       user's library.
    """

    def __init__(self, mendeley, token):
        if mendeley.client_secret:
            refresh_args = {'client_id': mendeley.client_id, 'client_secret': mendeley.client_secret}

            super(MendeleySession, self).__init__(client_id=mendeley.client_id,
                                                  token=token,
                                                  auto_refresh_url='%s/oauth/token' % mendeley.host,
                                                  auto_refresh_kwargs=refresh_args,
                                                  token_updater=lambda x: None)
        else:
            super(MendeleySession, self).__init__(client_id=mendeley.client_id, token=token)

        self.host = mendeley.host

        self.catalog = Catalog(self)
        self.documents = Documents(self, None)
        self.files = Files(self)
        self.groups = Groups(self)
        self.profiles = Profiles(self)
        self.trash = Trash(self, None)

    def group_members(self, group_id):
        return GroupMembers(self, group_id)

    def group_documents(self, group_id):
        return Documents(self, group_id)

    def group_trash(self, group_id):
        return Trash(self, group_id)

    def group_files(self, group_id):
        return Files(self, group_id=group_id)

    def document_files(self, document_id):
        return Files(self, document_id=document_id)

    def catalog_files(self, catalog_id):
        return Files(self, catalog_id=catalog_id)

    def request(self, method, url, data=None, headers=None, **kwargs):
        full_url = urljoin(self.host, url)

        if not headers:
            headers = {}

        headers['user-agent'] = self.__user_agent()

        rsp = super(MendeleySession, self).request(method, full_url, data, headers, **kwargs)

        if rsp.ok:
            return rsp
        else:
            raise MendeleyApiException(rsp)

    @staticmethod
    def __user_agent():
        return 'mendeley/%s %s/%s %s/%s' % (__version__,
                                            platform.python_implementation(),
                                            platform.python_version(),
                                            platform.system(),
                                            platform.release())


from setuptools import setup

__version__ = None
with open('mendeley/version.py') as f:
    exec(f.read())

setup(
    name='mendeley',
    version=__version__,
    packages=['mendeley', 'mendeley.models', 'mendeley.resources'],
    url='http://dev.mendeley.com',
    license='Apache',
    author='Mendeley',
    author_email='api@mendeley.com',
    description='Python SDK for the Mendeley API',

    install_requires=[
        'arrow==0.4.4',
        'future==0.14.1',
        'memoized-property==1.0.2',
        'requests==2.4.3',
        'requests-oauthlib==0.4.2',
        'oauthlib==0.6.3'
    ],

    tests_require=[
        'pytest==2.6.4',
        'vcrpy==1.1.2'
    ],

    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ]
)



from mendeley.models.profiles import Profile
from mendeley.resources.base import GetByIdResource


class Profiles(GetByIdResource):
    """
    Top-level resource for accessing profiles.
    """
    _url = '/profiles'

    def __init__(self, session):
        self.session = session

    def get(self, id):
        """
        Retrieves a profile by ID.

        :param id: the ID of the profile to get.
        :return: a :class:`Profile <mendeley.models.profiles.Profile>`.
        """
        return super(Profiles, self).get(id)

    @property
    def me(self):
        """
        The :class:`Profile <mendeley.models.profiles.Profile>` of the logged-in user.
        """
        return self.get('me')

    @property
    def _session(self):
        return self.session

    def _obj_type(self, **kwargs):
        return Profile



import arrow

from mendeley.response import ResponseObject


class Discipline(ResponseObject):
    """
    The discipline of a :class:`Profile <mendeley.models.profiles.Profile>`.

    .. attribute:: name
    .. attribute:: subdisciplines
    """
    @classmethod
    def fields(cls):
        return ['name', 'subdisciplines']


class Photo(ResponseObject):
    """
    A photo, associated with a :class:`Profile <mendeley.models.profiles.Profile>` or
    :class:`Group <mendeley.models.groups.Group>`.

    .. attribute:: original
    .. attribute:: standard
    .. attribute:: square
    """
    @classmethod
    def fields(cls):
        return ['original', 'standard', 'square']


class Location(ResponseObject):
    """
    A location, associated with a :class:`Profile <mendeley.models.profiles.Profile>`.

    .. attribute:: latitude
    .. attribute:: longitude
    .. attribute:: name
    """
    @classmethod
    def fields(cls):
        return ['latitude', 'longitude', 'name']


class Education(ResponseObject):
    """
    Education details, associated with a :class:`Profile <mendeley.models.profiles.Profile>`.

    .. attribute:: institution
    .. attribute:: degree
    .. attribute:: website
    """
    @property
    def start_date(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'start_date' in self.json:
            return arrow.get(self.json['start_date'])
        else:
            return

    @property
    def end_date(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'end_date' in self.json:
            return arrow.get(self.json['end_date'])
        else:
            return None

    @classmethod
    def fields(cls):
        return ['institution', 'degree', 'website']


class Employment(ResponseObject):
    """
    Employment details, associated with a :class:`Profile <mendeley.models.profiles.Profile>`.

    .. attribute:: institution
    .. attribute:: position
    .. attribute:: website
    .. attribute:: classes
    """
    @property
    def start_date(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'start_date' in self.json:
            return arrow.get(self.json['start_date'])
        else:
            return

    @property
    def end_date(self):
        """
        an :class:`Arrow <arrow.arrow.Arrow>` object.
        """
        if 'end_date' in self.json:
            return arrow.get(self.json['end_date'])
        else:
            return None

    @classmethod
    def fields(cls):
        return ['institution', 'position', 'website', 'classes']


class Person(ResponseObject):
    """
    A person, associated with a :class:`Document <mendeley.models.base_documents.BaseDocument>`, as either an author
    or an editor.

    .. attribute:: first_name
    .. attribute:: last_name
    """
    @classmethod
    def fields(cls):
        return ['first_name', 'last_name']

    @staticmethod
    def create(first_name, last_name):
        """
        Creates a person object, to be used when creating or updating a
        :class:`Document <mendeley.models.base_documents.BaseDocument>`.

        :param first_name:
        :param last_name:
        :return: a :class:`Person <mendeley.models.common.Person>`.
        """
        return Person({'first_name': first_name, 'last_name': last_name})



from mendeley.resources.base import ListResource, GetByIdResource


class DocumentsBase(GetByIdResource, ListResource):
    def __init__(self, session, group_id):
        self.session = session
        self.group_id = group_id

    def get(self, id, view=None):
        return super(DocumentsBase, self).get(id, view=view)

    def list(self, page_size=None, view=None, sort=None, order=None, modified_since=None, deleted_since=None):
        return super(DocumentsBase, self).list(page_size,
                                               view=view,
                                               sort=sort,
                                               order=order,
                                               modified_since=modified_since,
                                               deleted_since=deleted_since,
                                               group_id=self.group_id)

    def iter(self, page_size=None, view=None, sort=None, order=None, modified_since=None, deleted_since=None):
        return super(DocumentsBase, self).iter(page_size,
                                               view=view,
                                               sort=sort,
                                               order=order,
                                               modified_since=modified_since,
                                               deleted_since=deleted_since,
                                               group_id=self.group_id)

    @property
    def _session(self):
        return self.session

    def _obj_type(self, **kwargs):
        return self._view_type(kwargs.get('view'))

    @staticmethod
    def _view_type(view):
        raise NotImplementedError